# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import json
from collections.abc import Mapping
from typing import Any

import msgpack
import numpy as np
import pytest

from power_grid_model import DatasetType
from power_grid_model._core.dataset_definitions import ComponentType
from power_grid_model._core.utils import get_dataset_type, is_columnar, is_sparse
from power_grid_model.data_types import BatchDataset, Dataset, DenseBatchData, SingleComponentData, SingleDataset
from power_grid_model.enum import ComponentAttributeFilterOptions
from power_grid_model.utils import json_deserialize, json_serialize, msgpack_deserialize, msgpack_serialize


def to_json(data, raw_buffer: bool = False, indent: int | None = None):
    indent = None if indent is None or indent < 0 else indent
    separators = (",", ":") if indent is None else None
    result = json.dumps(data, indent=indent, separators=separators)

    if raw_buffer:
        return result.encode("utf-8")

    return result


def to_msgpack(data):
    return msgpack.packb(data)


def from_msgpack(data):
    return msgpack.unpackb(data)


def is_non_compact_list(data_entry):
    return isinstance(data_entry, dict)


def is_serialized_data_type_deducible(serialized_input, data_filter) -> bool:
    """Find out if deserialization of serialized_input would contain at least one component with row based data"""
    serialized_input_data = serialized_input["data"]
    if not serialized_input_data or isinstance(data_filter, ComponentAttributeFilterOptions):
        return False
    if data_filter is None:
        return True

    if not serialized_input["is_batch"]:
        components = set(serialized_input_data.keys())
    else:
        components = set()
        for scenario in serialized_input_data:
            components.update(scenario.keys())

    return any(data_filter[comp] is None for comp in components.intersection(data_filter.keys()))


def is_columnar_filter(data_filter, component) -> bool:
    """A function to find if a data_filter will give out row or columnar format for a component"""
    if data_filter is None:
        return False
    if isinstance(data_filter, ComponentAttributeFilterOptions):
        return True
    return data_filter[component] is not None


def is_attribute_filtered_out(data_filter, component, attribute) -> bool:
    """Checks if attribute is being filtered out / excluded"""
    if data_filter is None or isinstance(data_filter, ComponentAttributeFilterOptions):
        return False

    if component not in data_filter:
        return True

    return not (
        data_filter[component] is None
        or isinstance(data_filter[component], ComponentAttributeFilterOptions)
        or attribute in data_filter[component]
    )


def is_component_filtered_out(data_filter, component) -> bool:
    """Checks if component is being filtered out / excluded"""
    if data_filter is None or isinstance(data_filter, ComponentAttributeFilterOptions):
        return False
    return component not in data_filter


def is_sparse_data_input(serialized_input_data, component):
    """Checks if serialized_input_data will be of sparse format after deserialization.
    If there are uneven ids for scenarios in the serialized input, that would result in sparse data"""
    if not serialized_input_data["is_batch"]:
        raise ValueError("Checking if a non batch data is sparse")

    all_scenarios_ids: list[set[int]] = list(set())
    for scenario in serialized_input_data["data"]:
        scenario_ids = set()
        if component not in scenario:
            continue
        for scenario_comp_name_comp_idx in scenario[component]:
            if is_non_compact_list(scenario_comp_name_comp_idx):
                scenario_ids.add(scenario_comp_name_comp_idx["id"])
            else:
                id_attr_idx = serialized_input_data["attributes"][component].index("id")
                scenario_ids.add(scenario_comp_name_comp_idx[id_attr_idx])
        all_scenarios_ids.append(scenario_ids)
    first_scenario_ids = next(iter(all_scenarios_ids))

    return any(first_scenario_ids != scenario_ids for scenario_ids in all_scenarios_ids)


def empty_dataset(dataset_type: DatasetType = DatasetType.input):
    return {
        "version": "1.0",
        "type": dataset_type,
        "is_batch": False,
        "attributes": {},
        "data": {},
    }


def simple_input_dataset():
    return {
        "attributes": {},
        "data": {"node": [{"id": 5}], "source": [{"id": 6}, {"id": 7}]},
        "is_batch": False,
        "type": "input",
        "version": "1.0",
    }


def simple_asym_input_dataset():
    return {
        "attributes": {},
        "data": {"asym_load": [{"id": 5, "p_specified": [10.0, 20.0, 30.0]}]},
        "is_batch": False,
        "type": "input",
        "version": "1.0",
    }


def full_input_dataset():
    result = empty_dataset(DatasetType.input)
    result["attributes"] = {
        "node": ["id", "u_rated"],
        "sym_load": ["id", "node", "status", "type", "p_specified", "q_specified"],
        "source": ["id", "node", "status", "u_ref", "sk"],
    }
    result["data"] = {
        "node": [[1, 10.5e3], [2, 10.5e3], [3, 10.5e3]],
        "line": [
            {
                "id": 4,
                "from_node": 1,
                "to_node": 2,
                "from_status": 1,
                "to_status": 1,
                "r1": 0.11,
                "x1": 0.12,
                "c1": 4e-05,
                "tan1": 0.1,
                "i_n": 500.0,
            },
            {
                "id": 5,
                "from_node": 2,
                "to_node": 3,
                "from_status": 1,
                "to_status": 1,
                "r1": 0.15,
                "x1": 0.16,
                "c1": 5e-05,
                "tan1": 0.12,
                "i_n": 550.0,
            },
        ],
        "source": [
            [15, 1, 1, 1.03, 1e20],
            [16, 1, 1, 1.04, None],
            {
                "id": 17,
                "node": 1,
                "status": 1,
                "u_ref": 1.03,
                "sk": 1e10,
                "rx_ratio": 0.2,
            },
        ],
        "sym_load": [[7, 2, 1, 0, 1.01e6, 0.21e6], [8, 3, 1, 0, 1.02e6, 0.22e6]],
    }
    return result


def single_update_dataset():
    result = empty_dataset(DatasetType.update)
    result["attributes"] = {
        "sym_load": ["status", "p_specified", "q_specified"],
        "source": ["status"],
    }
    result["data"] = {
        "line": [{}, {"from_status": 1, "to_status": 1}],
        "source": [[1], [0], {"status": 1, "u_ref": 1.03}],
        "sym_load": [[9, 1.01e6, 0.21e6], [10, 1.02e6, 0.22e6]],
    }
    return result


def uniform_batch_update_dataset():
    return {
        "version": "1.0",
        "type": "update",
        "is_batch": True,
        "attributes": {
            "sym_load": ["id", "p_specified", "q_specified"],
            "asym_load": ["id", "p_specified"],
        },
        "data": [
            {"sym_load": [[7, 20.0, 50.0]], "asym_load": [[9, [100.0, None, 200.0]]]},
            {"sym_load": [[7, None, None]], "asym_load": [[9, None]]},
            {
                "sym_load": [{"id": 7, "status": 0}],
                "asym_load": [{"id": 9, "q_specified": [70.0, 80.0, 90.0]}],
            },
        ],
    }


def inhomogeneous_batch_update_dataset():
    return {
        "version": "1.0",
        "type": "update",
        "is_batch": True,
        "attributes": {
            "sym_load": ["id", "p_specified", "q_specified"],
            "asym_load": ["id", "p_specified"],
        },
        "data": [
            {"sym_load": [[7, 20.0, 50.0]], "asym_load": [[9, [100.0, None, 200.0]]]},
            {"asym_load": [[9, None]]},
            {
                "sym_load": [[7, None, 10.0], {"id": 8, "status": 0}],
                "asym_load": [{"id": 9, "q_specified": [70.0, 80.0, 90.0]}],
            },
        ],
    }


def sparse_batch_update_dataset():
    return {
        "version": "1.0",
        "type": "update",
        "is_batch": True,
        "attributes": {},
        "data": [
            {"sym_load": [{"id": 7, "q_specified": 50.0}]},
            {"sym_load": [{"id": 8, "q_specified": 33.333333333333336}]},
            {
                "sym_load": [
                    {"id": 7, "q_specified": 10.0},
                    {"id": 8, "q_specified": 2.5},
                ]
            },
        ],
    }


def single_sym_output_dataset():
    result = empty_dataset(DatasetType.sym_output)
    result["data"] = {
        "node": [
            {
                "id": 1,
                "energized": 1,
                "u_pu": 1.01,
                "u_angle": 0.21,
                "u": 1.02e3,
                "p": 1.01e6,
                "q": 4.1e5,
            }
        ]
    }
    return result


def batch_sym_output_dataset():
    result = empty_dataset(DatasetType.sym_output)
    result["is_batch"] = True
    result["data"] = [
        {
            "node": [
                {
                    "id": 1,
                    "energized": 1,
                    "u_pu": 1.01,
                    "u_angle": 0.21,
                    "u": 1.02e3,
                    "p": 1.01e6,
                    "q": 4.1e5,
                }
            ]
        },
        {
            "node": [
                {
                    "id": 1,
                    "energized": 0,
                    "u_pu": 0.0,
                    "u_angle": 0.0,
                    "u": 0.0,
                    "p": 0.0,
                    "q": 0.0,
                }
            ]
        },
    ]
    return result


def single_asym_output_dataset():
    result = empty_dataset(DatasetType.asym_output)
    result["data"] = {
        "node": [
            {
                "id": 1,
                "energized": 1,
                "u_pu": [1.01, 1.06, 0.99],
                "u_angle": [0.21, 2.01, 4.1],
                "u": [1.02e3, 1.07e3, 9.9e2],
                "p": [1.01e6, 1.03e6, 9.8e5],
                "q": [4.1e5, 4.2e5, 4.0e5],
            }
        ]
    }
    return result


def single_sc_output_dataset():
    result = empty_dataset(DatasetType.sc_output)
    result["attributes"] = {"fault": ["id", "i_f"]}
    result["data"] = {
        "node": [
            {
                "id": 1,
                "energized": 1,
                "u_pu": [1.01, 1.06, 0.99],
                "u_angle": [0.21, 2.01, 4.1],
                "u": [1.02e3, 1.07e3, 9.9e2],
            }
        ],
        "fault": [{"id": 1, "i_f": [3.0e3, 2.0e3, 3.4e3]}],
    }
    return result


@pytest.fixture(
    params=[
        empty_dataset,
        simple_input_dataset,
        full_input_dataset,
        single_update_dataset,
        uniform_batch_update_dataset,
        inhomogeneous_batch_update_dataset,
        sparse_batch_update_dataset,
        single_sym_output_dataset,
        batch_sym_output_dataset,
        single_asym_output_dataset,
        single_sc_output_dataset,
        simple_asym_input_dataset,
    ]
)
def serialized_data(request):
    return request.param()


@pytest.fixture(
    params=[
        pytest.param(None, id="All row filter"),
        pytest.param(ComponentAttributeFilterOptions.everything, id="All columnar filter"),
        pytest.param(ComponentAttributeFilterOptions.relevant, id="All relevant columnar filter"),
        pytest.param({ComponentType.node: ["id"], "sym_load": ["id"]}, id="columnar filter"),
        pytest.param({ComponentType.node: ["id"], "sym_load": None}, id="mixed columnar/row filter"),
        pytest.param({ComponentType.node: ["id"], "shunt": None}, id="unused component filter"),
        pytest.param(
            {
                ComponentType.node: ["id"],
                "line": ComponentAttributeFilterOptions.everything,
                "sym_load": None,
                "asym_load": ComponentAttributeFilterOptions.relevant,
            },
            id="mixed filter",
        ),
    ]
)
def data_filters(request):
    """Data filers for deserialization.
    Note: List / set is not supported for testing since it can be replicated using dict."""
    return request.param


def split_deserialized_dataset_into_individual_scenario(scenario_idx, deserialized_dataset):
    deserialized_scenario = {}
    for component, component_values in deserialized_dataset.items():
        if not is_sparse(component_values):
            if not is_columnar(component_values):
                deserialized_scenario[component] = component_values[scenario_idx]
            else:
                deserialized_scenario[component] = {}
                for attr, attr_values in component_values.items():
                    deserialized_scenario[component][attr] = attr_values[scenario_idx]
        else:
            scenario_slice = slice(
                component_values["indptr"][scenario_idx],
                component_values["indptr"][scenario_idx + 1],
            )
            if not is_columnar(component_values):
                sparse_component_data = component_values["data"][scenario_slice]
                component_in_scenario = len(sparse_component_data) > 0
            else:
                sparse_component_data = {
                    attr: attr_value[scenario_slice] for attr, attr_value in component_values["data"].items()
                }
                component_in_scenario = len(next(iter(sparse_component_data.values()))) > 0

            if component_in_scenario:
                deserialized_scenario[component] = sparse_component_data
    return deserialized_scenario


def assert_not_a_value(value: np.ndarray):
    """Assert value contains only NaN-like values.

    Depending on type, maps to np.nan, -127, na_IntID, etc.
    """
    if value.dtype.names:
        for key in value.dtype.names:
            assert_not_a_value(value[key])
    elif np.issubdtype(value.dtype, np.integer):
        na_int = np.finfo(value.dtype).min
        assert np.all(value == na_int)
    else:
        assert np.all(np.isnan(value))


def assert_almost_equal(value: np.ndarray, reference: Any):
    """Recursively walk through the values and references, assert almost equal.

    Correctly handle:
      * Both the sparse and dense data representations
      * None in the reference type maps to not a value.
    """
    if isinstance(reference, list):
        assert len(value) == len(reference)
        for v, r in zip(value, reference):
            assert_almost_equal(v, r)
    elif isinstance(reference, dict):
        assert len(value) == len(reference)
        assert value.dtype.names is not None
        for attribute in value.dtype.names:
            if attribute in reference:
                assert_almost_equal(value[attribute], reference[attribute])
            else:
                assert_not_a_value(reference[attribute])
    elif reference is None:
        assert_not_a_value(value)
    else:
        np.testing.assert_almost_equal(value, reference)


def assert_single_dataset_entries(
    deserialized_dataset: SingleDataset,
    serialized_dataset: Mapping[str, Any],
    data_filter,
):
    """Check a SingleDataset's individual components for correctness"""
    for component, serialized_input in serialized_dataset["data"].items():
        if is_component_filtered_out(data_filter, component):
            assert component not in deserialized_dataset
            continue

        deserialized_output = deserialized_dataset[component]

        # Complete component data checks
        if is_columnar_filter(data_filter, component):
            assert isinstance(deserialized_output, dict)
            # Assert number of batches are equal for all attributes
            assert all(len(v) == len(serialized_input) for v in deserialized_output.values())
            # Assert number of attributes are not more than input for all batches
            # Discrepencies in equal / less than input captured in individual check
            # TODO Test for deserialize only relevant attributes
            # assert all(len(v) >= len(deserialized_output) for v in serialized_input)
        else:
            assert isinstance(deserialized_output, np.ndarray)
            assert len(deserialized_output) == len(serialized_input)

        assert_individual_data_entry(
            serialized_dataset,
            data_filter,
            component,
            serialized_input,
            deserialized_output,
        )


def assert_individual_data_entry(serialized_dataset, data_filter, component, serialized_input, deserialized_output):
    """Checks each data entry of the component for correctness"""
    for comp_idx, input_entry in enumerate(serialized_input):
        if is_non_compact_list(input_entry):
            for attr in input_entry:
                if is_columnar_filter(data_filter, component):
                    if is_attribute_filtered_out(data_filter, component, attr):
                        assert attr not in deserialized_output
                        continue
                    assert attr in deserialized_output
                    assert_almost_equal(
                        deserialized_output[attr][comp_idx],
                        input_entry[attr],
                    )
                else:
                    assert attr in deserialized_output[comp_idx].dtype.names
                    assert_almost_equal(
                        deserialized_output[comp_idx][attr],
                        input_entry[attr],
                    )
        else:
            assert component in serialized_dataset["attributes"]
            for attr_idx, attr in enumerate(serialized_dataset["attributes"][component]):
                if is_columnar_filter(data_filter, component):
                    if is_attribute_filtered_out(data_filter, component, attr):
                        assert attr not in deserialized_output
                        continue
                    assert attr in deserialized_output
                    assert_almost_equal(
                        deserialized_output[attr][comp_idx],
                        input_entry[attr_idx],
                    )
                else:
                    assert attr in deserialized_output[comp_idx].dtype.names
                    assert_almost_equal(
                        deserialized_output[comp_idx][attr],
                        input_entry[attr_idx],
                    )


def assert_single_dataset_structure(deserialized_dataset, data_filter):
    for component, component_result in deserialized_dataset.items():
        if is_columnar_filter(data_filter, component):
            for attr_name, arr in component_result.items():
                assert isinstance(arr, np.ndarray)
                assert isinstance(attr_name, str)
                assert len(arr.shape) in [1, 2]
        else:
            assert isinstance(component_result, np.ndarray)
            assert component_result.ndim == 1


def assert_batch_dataset_structure(
    deserialized_dataset: BatchDataset,
    serialized_dataset: Mapping[str, Any],
    data_filter,
):
    """Checks if the structure of the batch dataset is correct.
    Then splits into individual scenario's dataset and checks if all of them are correct."""
    batch_dataset_ndim = 2

    # Check structure of the whole BatchDataset
    assert isinstance(serialized_dataset["data"], list)
    for component, component_values in deserialized_dataset.items():
        if is_sparse_data_input(serialized_dataset, component):
            component_indptr = component_values["indptr"]
            component_data = component_values["data"]
            assert isinstance(component_indptr, np.ndarray)
            assert len(component_indptr) == len(serialized_dataset["data"]) + 1
            if is_columnar_filter(data_filter, component):
                assert isinstance(component_data, dict)
                for attr, attr_value in component_data.items():
                    assert isinstance(attr, str)
                    assert isinstance(attr_value, np.ndarray)
                    assert attr_value.ndim == 1
            else:
                assert isinstance(component_data, np.ndarray)
                assert component_data.ndim == 1
                assert len(component_data) == component_indptr[-1]
        elif is_columnar_filter(data_filter, component):
            for attr, attr_value in component_values.items():
                assert isinstance(attr, str)
                assert isinstance(attr_value, np.ndarray)
                assert len(attr_value.shape) in [2, 3]
                assert len(attr_value) == len(serialized_dataset["data"])
        else:
            assert isinstance(component_values, np.ndarray)
            assert len(component_values.shape) == batch_dataset_ndim
            assert len(component_values) == len(serialized_dataset["data"])


def assert_serialization_correct(deserialized_dataset: Dataset, serialized_dataset: Mapping[str, Any], data_filter):
    """Assert the dataset correctly reprensents the input data."""
    if serialized_dataset["is_batch"]:
        assert_batch_dataset_structure(deserialized_dataset, serialized_dataset, data_filter=data_filter)

        # Split into individual SingleDataset and check if they all are correct
        for scenario_idx, scenario in enumerate(serialized_dataset["data"]):
            serialized_scenario = {k: v for k, v in serialized_dataset.items() if k not in ("data", "is_batch")}
            serialized_scenario["is_batch"] = False
            serialized_scenario["data"] = scenario

            deserialized_scenario = split_deserialized_dataset_into_individual_scenario(
                scenario_idx, deserialized_dataset
            )
            assert_single_dataset_entries(deserialized_scenario, serialized_scenario, data_filter=data_filter)
    else:
        assert_single_dataset_structure(deserialized_dataset, data_filter)

        assert_single_dataset_entries(
            deserialized_dataset,
            serialized_dataset,
            data_filter=data_filter,
        )


def _check_only_relevant_attributes_present(component_values: SingleComponentData | DenseBatchData) -> bool:
    if isinstance(component_values, np.ndarray):
        return True
    for array in component_values.values():
        if (array.dtype == np.float64 and np.isnan(array).all()) or (
            array.dtype in (np.int32, np.int8) and np.all(array == np.iinfo(array.dtype).min)
        ):
            return False
    return True


def assert_deserialization_filtering_correct(deserialized_dataset: Dataset, data_filter) -> bool:
    if data_filter is ComponentAttributeFilterOptions.everything:
        return True
    if data_filter is ComponentAttributeFilterOptions.relevant:
        for component_values in deserialized_dataset.values():
            buffer = component_values if not is_sparse(component_values) else component_values["data"]
            if not _check_only_relevant_attributes_present(buffer):
                return False
    return True


@pytest.mark.parametrize("raw_buffer", [True, False])
def test_json_deserialize_data(serialized_data, data_filters, raw_buffer: bool):
    data = to_json(serialized_data, raw_buffer=raw_buffer)
    result = json_deserialize(data, data_filter=data_filters)

    assert isinstance(result, dict)
    assert_serialization_correct(result, serialized_data, data_filters)

    if is_serialized_data_type_deducible(serialized_data, data_filter=data_filters):
        result_type = get_dataset_type(result)
        assert result_type == serialized_data["type"]


def test_msgpack_deserialize_data(serialized_data, data_filters):
    data = to_msgpack(serialized_data)
    result = msgpack_deserialize(data, data_filters)

    assert isinstance(result, dict)
    assert_serialization_correct(result, serialized_data, data_filter=data_filters)

    if is_serialized_data_type_deducible(serialized_data, data_filter=data_filters):
        result_type = get_dataset_type(result)
        assert result_type == serialized_data["type"]


@pytest.mark.parametrize(
    "dataset_type",
    [
        DatasetType.input,
        DatasetType.update,
        DatasetType.sym_output,
        DatasetType.asym_output,
        DatasetType.sc_output,
    ],
)
@pytest.mark.parametrize(
    "use_compact_list",
    [pytest.param(True, id="compact_list"), pytest.param(False, id="non-compact_list")],
)
def test_json_serialize_empty_dataset(dataset_type, use_compact_list: bool):
    for indent in (-1, 0, 2, 4):
        reference = to_json(empty_dataset(dataset_type), indent=indent)
        assert isinstance(reference, str)

        result = json_serialize({}, dataset_type, use_compact_list=use_compact_list, indent=indent)
        assert isinstance(result, str)
        assert result == reference

        with pytest.raises(ValueError, match="At least one component should have row based data"):
            json_serialize({}, use_compact_list=use_compact_list, indent=indent)


@pytest.mark.parametrize(
    "dataset_type",
    [
        DatasetType.input,
        DatasetType.update,
        DatasetType.sym_output,
        DatasetType.asym_output,
        DatasetType.sc_output,
    ],
)
@pytest.mark.parametrize(
    "use_compact_list",
    [pytest.param(True, id="compact_list"), pytest.param(False, id="non-compact_list")],
)
def test_msgpack_serialize_empty_dataset(dataset_type, use_compact_list):
    reference = empty_dataset(dataset_type)
    assert from_msgpack(msgpack_serialize({}, dataset_type, use_compact_list=use_compact_list)) == reference

    with pytest.raises(ValueError, match="At least one component should have row based data"):
        json_serialize({}, use_compact_list=use_compact_list)


@pytest.mark.parametrize(
    ("deserialize", "serialize", "pack"),
    [
        pytest.param(json_deserialize, json_serialize, to_json, id="json"),
        pytest.param(msgpack_deserialize, msgpack_serialize, to_msgpack, id="msgpack"),
    ],
)
def test_serialize_deserialize_type_deduction(deserialize, serialize, serialized_data, data_filters, pack):
    deserialized_data = deserialize(pack(serialized_data), data_filter=data_filters)
    full_result = serialize(deserialized_data, serialized_data["type"])

    if is_serialized_data_type_deducible(serialized_data, data_filter=data_filters):
        assert serialize(deserialized_data) == full_result
    else:
        with pytest.raises(ValueError, match="The dataset type could not be deduced"):
            serialize(deserialized_data)


@pytest.mark.parametrize(
    ("deserialize", "serialize", "pack"),
    [
        pytest.param(json_deserialize, json_serialize, to_json, id="json"),
        pytest.param(msgpack_deserialize, msgpack_serialize, to_msgpack, id="msgpack"),
    ],
)
def test_serialize_deserialize_double_round_trip(deserialize, serialize, serialized_data, data_filters, pack):
    """
    Repeated deserialization and serialization must result in the same deserialized data and serialization string.
    """
    test_data = pack(serialized_data)

    deserialized_result_a = deserialize(test_data, data_filters)
    serialized_result_a = serialize(deserialized_result_a, dataset_type=serialized_data["type"])

    deserialized_result_b = deserialize(serialized_result_a, data_filters)
    serialized_result_b = serialize(deserialized_result_b, dataset_type=serialized_data["type"])

    assert serialized_result_a == serialized_result_b
    assert list(deserialized_result_b) == list(deserialized_result_a)
    assert assert_deserialization_filtering_correct(deserialized_result_a, data_filters)
    assert assert_deserialization_filtering_correct(deserialized_result_b, data_filters)

    for (component_a, component_result_a), component_result_b in zip(
        deserialized_result_a.items(), deserialized_result_b.values()
    ):
        is_non_uniform = is_sparse(component_result_a)
        assert is_non_uniform == is_sparse(component_result_b)

        if not is_non_uniform:
            component_data_a = component_result_a
            component_data_b = component_result_b
        else:
            np.testing.assert_array_equal(component_result_a["indptr"], component_result_b["indptr"])
            component_data_a = component_result_a["data"]
            component_data_b = component_result_b["data"]

        if is_columnar_filter(data_filters, component_a):
            assert component_data_a.keys() == component_data_b.keys()
            assert all(v_a.dtype == v_b.dtype for v_a, v_b in zip(component_data_a.values(), component_data_b.values()))
            fields_or_keys = component_data_a.keys()
        else:
            assert component_data_a.dtype == component_data_b.dtype
            fields_or_keys = component_data_a.dtype.names

        for field in fields_or_keys:
            field_result_a = component_data_a[field]
            field_result_b = component_data_b[field]

            nan_a = np.isnan(field_result_a)
            nan_b = np.isnan(field_result_b)

            np.testing.assert_array_equal(nan_a, nan_b)
            np.testing.assert_allclose(field_result_a[~nan_a], field_result_b[~nan_b], rtol=1e-15)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""
Utilities used for validation. Only errors_to_string() is intended for end users.
"""

import re
from typing import Any, cast

import numpy as np

from power_grid_model._core.dataset_definitions import ComponentType, DatasetType
from power_grid_model._core.power_grid_meta import power_grid_meta_data
from power_grid_model.data_types import SingleArray, SingleComponentData, SingleDataset
from power_grid_model.validation.errors import ValidationError


def _eval_expression(data: np.ndarray, expression: int | float | str) -> np.ndarray:
    """
    Wrapper function that checks the type of the 'expression'. If the expression is a string, it is assumed to be a
    field expression and the expression is validated. Otherwise it is assumed to be a numerical value and the value
    is casted to a numpy 'array'.

    Args:
        data: A numpy structured array
        expression: A numerical value, or a string, representing a (combination of) field(s)

    Returns: The number or an evaluation of the field name(s) in the data, always represented as a Numpy 'array'.

    Examples:
        123 -> np.array(123)
        123.4 -> np.array(123.4)
        'value' -> data['value']
        'foo/bar' -> data['foo'] / data[bar]

    """
    if isinstance(expression, str):
        return _eval_field_expression(data, expression)
    return np.array(expression)


def _eval_field_expression(data: np.ndarray, expression: str) -> np.ndarray:
    """
    A field expression can either be the name of a field (e.g. 'field_x') in the data, or a ratio between two fields
    (e.g. 'field_x / field_y'). The expression is checked on validity and then the fields are checked to be present in
    the data. If the expression is a single field name, the field is returned. If it is a ratio, the ratio is
    calculated and returned. Values divided by 0 will result in nan values without warning.

    Args:
        data: A numpy structured array
        expression: A string, representing a (combination of) field(s)

    Expression should be a combination of:
      - field names (may contain lower case letters, numbers and underscores)
      - a single mathematical operator /

    Returns: An evaluation of the field name(s) in the data.

    Examples:
        'value' -> data['value']
        'foo/bar' -> data['foo'] / data['bar']

    """
    # Validate the expression
    match = re.fullmatch(r"[a-z][a-z0-9_]*(\s*/\s*[a-z][a-z0-9_]*)?", expression)
    if not match:
        raise ValueError(f"Invalid field expression '{expression}'")

    if data.dtype.names is None:
        raise ValueError("No attributes available in meta")

    # Find all field names and check if they exist in the dataset
    fields = [f.strip() for f in expression.split("/")]
    for field in fields:
        if field not in data.dtype.names:
            raise KeyError(f"Invalid field name {field}")

    if len(fields) == 1:
        return data[fields[0]]

    max_num_fields = 2
    if len(fields) != max_num_fields:
        raise ValueError(f"There should be exactly {max_num_fields} fields, got {len(fields)} fields")
    zero_div = np.logical_or(np.equal(data[fields[1]], 0.0), np.logical_not(np.isfinite(data[fields[1]])))
    if np.any(zero_div):
        result = np.full_like(data[fields[0]], np.nan)
        np.true_divide(data[fields[0]], data[fields[1]], out=result, where=~zero_div)
        return result
    return np.true_divide(data[fields[0]], data[fields[1]])


def _update_input_data(input_data: SingleDataset, update_data: SingleDataset):
    """
    Update the input data using the available non-nan values in the update data.
    """

    merged_data = {component: array.copy() for component, array in input_data.items()}
    for component in update_data:
        _update_component_data(component, merged_data[component], update_data[component])
    return merged_data


def _update_component_data(
    component: ComponentType, input_data: SingleComponentData, update_data: SingleComponentData
) -> None:
    """
    Update the data in a single component data set, with another single component data set,
    indexed on the "id" field and only non-NaN values are overwritten.
    """
    if isinstance(input_data, np.ndarray) and isinstance(update_data, np.ndarray):
        return _update_component_array_data(component=component, input_data=input_data, update_data=update_data)

    raise NotImplementedError  # TODO(mgovers): add support for columnar data


def _update_component_array_data(component: ComponentType, input_data: SingleArray, update_data: SingleArray) -> None:
    """
    Update the data in a numpy array, with another numpy array,
    indexed on the "id" field and only non-NaN values are overwritten.
    """
    batch_dataset_ndim = 2
    if update_data.dtype.names is None:
        raise ValueError("Invalid data format")

    optional_ids_active = (
        "id" in update_data.dtype.names
        and np.all(update_data["id"] == np.iinfo(update_data["id"].dtype).min)
        and len(update_data["id"]) == len(input_data["id"])
    )
    update_data_ids = input_data["id"] if optional_ids_active else update_data["id"]

    for field in update_data.dtype.names:
        if field == "id":
            continue
        nan = _nan_type(component, field, DatasetType.update)
        mask = ~np.isnan(update_data[field]) if np.isnan(nan) else np.not_equal(update_data[field], nan)

        if mask.ndim == batch_dataset_ndim:
            for phase in range(mask.shape[1]):
                # find indexers of to-be-updated object
                sub_mask = mask[:, phase]
                idx = _get_indexer(input_data["id"], update_data_ids[sub_mask])
                # update
                input_data[field][idx, phase] = update_data[field][sub_mask, phase]
        else:
            # find indexers of to-be-updated object
            idx = _get_indexer(input_data["id"], update_data_ids[mask])
            # update
            input_data[field][idx] = update_data[field][mask]


def errors_to_string(
    errors: list[ValidationError] | dict[int, list[ValidationError]] | None,
    name: str = "the data",
    details: bool = False,
    id_lookup: list[str] | dict[int, str] | None = None,
) -> str:
    """
    Convert a set of errors (list or dict) to a human readable string representation.

    Args:
        errors: The error objects. List for input_data only, dict for batch data.
        name: Human understandable name of the dataset, e.g. input_data, or update_data.
        details: Display object ids and error specific information.
        id_lookup: A list or dict (int->str) containing textual object ids

    Returns:
        A human readable string representation of a set of errors.
    """
    if errors is None or len(errors) == 0:
        return f"{name}: OK"
    if isinstance(errors, dict):
        return "\n".join(errors_to_string(err, f"{name}, batch #{i}", details) for i, err in sorted(errors.items()))
    if len(errors) == 1 and not details:
        return f"There is a validation error in {name}:\n\t{errors[0]}"
    if len(errors) == 1:
        msg = f"There is a validation error in {name}:\n"
    else:
        msg = f"There are {len(errors)} validation errors in {name}:\n"
    if details:
        for error in errors:
            msg += "\n\t" + str(error) + "\n"
            msg += "".join(f"\t\t{k}: {v}\n" for k, v in error.get_context(id_lookup).items())
    else:
        msg += "\n".join(f"{i + 1:>4}. {err}" for i, err in enumerate(errors))
    return msg


def _nan_type(component: ComponentType, field: str, data_type: DatasetType = DatasetType.input):
    """Helper function to retrieve the nan value for a certain field as defined in the power_grid_meta_data."""
    return power_grid_meta_data[data_type][component].nans[field]


def _get_indexer(source: np.ndarray, target: np.ndarray, default_value: int | None = None) -> np.ndarray:
    """
    Given array of values from a source and a target dataset.
    Find the position of each value in the target dataset in the context of the source dataset.
    This is needed to update values in the dataset by id lookup.
    Internally this is done by sorting the input ids, then using binary search lookup.

    E.g.: Find the position of each id in an update (target) dataset in the input (source) dataset

    >>> input_ids = [1, 2, 3, 4, 5]
    >>> update_ids = [3]
    >>> assert _get_indexer(input_ids, update_ids) == np.array([2])

    Args:
        source: array of values in the source dataset
        target: array of values in the target dataset
        default_value (optional): the default index to provide for target values not in source

    Returns:
        np.ndarray: array of positions of the values from target dataset in the source dataset
            if default_value is None, (source[result] == target)
            else, ((source[result] == target) | (source[result] == default_value))

    Raises:
        IndexError: if default_value is None and there were values in target that were not in source
    """

    permutation_sort = np.argsort(source)  # complexity O(N_input * logN_input)
    indices = np.searchsorted(source, target, sorter=permutation_sort)  # complexity O(N_update * logN_input)

    if default_value is None:
        return permutation_sort[indices]

    if len(source) == 0:
        return np.full_like(target, fill_value=default_value)

    clipped_indices = np.take(permutation_sort, indices, mode="clip")
    return np.where(source[clipped_indices] == target, permutation_sort[clipped_indices], default_value)


def _set_default_value(
    data: SingleDataset, component: ComponentType, field: str, default_value: int | float | np.ndarray
):
    """
    This function sets the default value in the data that is to be validated, so the default values are included in the
    validation.

    Args:
        data: The input/update data set for all components
        component: The component of interest
        field: The field of interest
        default_value: Some values are not required, but will receive a default value in the C++ core. To do a proper
        input validation, these default values should be included in the validation. It can be a fixed value for the
        entire column (int/float) or be different for each element (np.ndarray).

    Returns:

    """
    if np.isnan(_nan_type(component, field)):
        mask = np.isnan(data[component][field])
    else:
        mask = data[component][field] == _nan_type(component, field)
    if isinstance(default_value, np.ndarray):
        data[component][field][mask] = default_value[mask]
    else:
        data[component][field][mask] = default_value


def _get_valid_ids(data: SingleDataset, ref_components: ComponentType | list[ComponentType]) -> list[int]:
    """
    This function returns the valid IDs specified by all ref_components.

    Args:
        data: The input/update data set for all components
        ref_components: The component or components in which we want to look for ids

    Returns:
        list[int]: the list of valid IDs
    """
    # For convenience, ref_component may be a string and we'll convert it to a 'list' containing that string as it's
    # single element.
    if isinstance(ref_components, (str, ComponentType)):
        ref_components = cast(list[ComponentType], [ref_components])

    # Create a set of ids by chaining the ids of all ref_components
    valid_ids = set()
    for ref_component in ref_components:
        if ref_component in data:
            nan = _nan_type(ref_component, "id")
            if np.isnan(nan):
                mask = ~np.isnan(data[ref_component]["id"])
            else:
                mask = np.not_equal(data[ref_component]["id"], nan)
            valid_ids.update(data[ref_component]["id"][mask])

    return list(valid_ids)


def _get_mask(data: SingleDataset, component: ComponentType, field: str, **filters: Any) -> np.ndarray:
    """
    Get a mask based on the specified filters. E.g. measured_terminal_type=MeasuredTerminalType.source.

    Args:
        data: The input/update data set for all components
        component: The component of interest
        field: The field of interest
        ref_components: The component or components in which we want to look for ids
        **filters: One or more filters on the dataset. E.

    Returns:
        np.ndarray: the mask
    """
    values = data[component][field]
    mask = np.ones(shape=values.shape, dtype=bool)
    for filter_field, filter_value in filters.items():
        mask = np.logical_and(mask, data[component][filter_field] == filter_value)

    return mask



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from pathlib import Path

import pytest

from power_grid_model._core.power_grid_model import PowerGridModel
from power_grid_model._core.utils import convert_batch_dataset_to_batch_list
from power_grid_model.utils import import_json_data
from tests.unit.utils import compare_result


@pytest.fixture
def deprecated_format_test_case():
    return Path(__file__).parent / "data/dummy-test-batch-dependent-not-cacheable"


def test_end_to_end__deprecated_serialization_format_single(deprecated_format_test_case):
    with pytest.deprecated_call():
        input_data = import_json_data(deprecated_format_test_case / "input.json", "input")
        sym_output = import_json_data(deprecated_format_test_case / "sym_output.json", "sym_output")

    model = PowerGridModel(input_data, system_frequency=50.0)
    result = model.calculate_power_flow()

    compare_result(result, sym_output, 1e-5, 1e-5)


def test_end_to_end__deprecated_serialization_format_batch(deprecated_format_test_case):
    with pytest.deprecated_call():
        input_data = import_json_data(deprecated_format_test_case / "input.json", "input")
        update = import_json_data(deprecated_format_test_case / "update_batch.json", "update")
        sym_output = import_json_data(deprecated_format_test_case / "sym_output_batch.json", "sym_output")

    model = PowerGridModel(input_data, system_frequency=50.0)
    result = model.calculate_power_flow(update_data=update)

    for scenario_result, scenario_expected in zip(
        convert_batch_dataset_to_batch_list(result), convert_batch_dataset_to_batch_list(sym_output)
    ):
        compare_result(scenario_result, scenario_expected, 1e-5, 1e-5)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from copy import copy, deepcopy

import numpy as np
import pytest

from power_grid_model import (
    ComponentAttributeFilterOptions,
    ComponentType,
    DatasetType,
    PowerGridModel,
    initialize_array,
)
from power_grid_model._core.utils import compatibility_convert_row_columnar_dataset
from power_grid_model.data_types import BatchDataset
from power_grid_model.errors import InvalidCalculationMethod, IterationDiverge, PowerGridBatchError, PowerGridError
from power_grid_model.utils import get_dataset_scenario
from power_grid_model.validation import assert_valid_input_data

from .utils import compare_result

"""
Testing network

source_1(1.0 p.u., 100.0 V) --internal_impedance(j10.0 ohm, sk=1000.0 VA, rx_ratio=0.0)--
-- node_0 (100.0 V) --load_2(const_i, -j5.0A, 0.0 W, 500.0 var)

u0 = 100.0 V - (j10.0 ohm * -j5.0 A) = 50.0 V

update_0:
    u_ref = 0.5 p.u. (50.0 V)
    q_specified = 100 var (-j1.0A)
u0 = 50.0 V - (j10.0 ohm * -j1.0 A) = 40.0 V

update_1:
    q_specified = 300 var (-j3.0A)
u0 = 100.0 V - (j10.0 ohm * -j3.0 A) = 70.0 V
"""


@pytest.fixture
def input_row():
    node = initialize_array(DatasetType.input, ComponentType.node, 1)
    node["id"] = 0
    node["u_rated"] = 100.0

    source = initialize_array(DatasetType.input, ComponentType.source, 1)
    source["id"] = 1
    source["node"] = 0
    source["status"] = 1
    source["u_ref"] = 1.0
    source["sk"] = 1000.0
    source["rx_ratio"] = 0.0

    sym_load = initialize_array(DatasetType.input, ComponentType.sym_load, 1)
    sym_load["id"] = 2
    sym_load["node"] = 0
    sym_load["status"] = 1
    sym_load["type"] = 2
    sym_load["p_specified"] = 0.0
    sym_load["q_specified"] = 500.0

    return {
        ComponentType.node: node,
        ComponentType.source: source,
        ComponentType.sym_load: sym_load,
    }


@pytest.fixture
def input_col(input_row):
    return compatibility_convert_row_columnar_dataset(
        input_row, ComponentAttributeFilterOptions.relevant, DatasetType.input
    )


@pytest.fixture(params=["input_row", "input_col"])
def input(request):
    return request.getfixturevalue(request.param)


@pytest.fixture
def sym_output():
    node = initialize_array(DatasetType.sym_output, ComponentType.node, 1)
    node["id"] = 0
    node["u"] = 50.0
    node["u_pu"] = 0.5
    node["u_angle"] = 0.0

    return {ComponentType.node: node}


@pytest.fixture
def update_batch_row():
    source = initialize_array(DatasetType.update, ComponentType.source, 1)
    source["id"] = 1
    source["u_ref"] = 0.5

    sym_load = initialize_array(DatasetType.update, ComponentType.sym_load, 2)
    sym_load["id"] = [2, 2]
    sym_load["q_specified"] = [100.0, 300.0]

    return {
        ComponentType.source: {
            "data": source,
            "indptr": np.array([0, 1, 1]),
        },
        ComponentType.sym_load: {
            "data": sym_load,
            "indptr": np.array([0, 1, 2]),
        },
    }


@pytest.fixture
def update_batch_col(update_batch_row):
    return compatibility_convert_row_columnar_dataset(
        update_batch_row, ComponentAttributeFilterOptions.relevant, DatasetType.update
    )


@pytest.fixture(params=["update_batch_row", "update_batch_col"])
def update_batch(request):
    return request.getfixturevalue(request.param)


@pytest.fixture
def sym_output_batch():
    node = initialize_array(DatasetType.sym_output, ComponentType.node, (2, 1))
    node["id"] = [[0], [0]]
    node["u"] = [[40.0], [70.0]]
    node["u_pu"] = [[0.4], [0.7]]
    node["u_angle"] = [[0.0], [0.0]]

    return {
        ComponentType.node: node,
    }


@pytest.fixture
def model(input):
    return PowerGridModel(input)


@pytest.fixture
def empty_model():
    return PowerGridModel({})


def test_simple_power_flow(model: PowerGridModel, sym_output):
    result = model.calculate_power_flow()
    compare_result(result, sym_output, rtol=0.0, atol=1e-8)


def test_simple_permanent_update(model: PowerGridModel, update_batch, sym_output_batch):
    model.update(update_data=get_dataset_scenario(update_batch, 0))  # single permanent model update
    result = model.calculate_power_flow()
    expected_result = get_dataset_scenario(sym_output_batch, 0)
    compare_result(result, expected_result, rtol=0.0, atol=1e-8)


def test_update_error(model: PowerGridModel):
    load_update = initialize_array(DatasetType.update, ComponentType.sym_load, 1)
    load_update["id"] = 5
    update_data = {ComponentType.sym_load: load_update}
    with pytest.raises(PowerGridError, match="The id cannot be found:"):
        model.update(update_data=update_data)
    update_data_col = compatibility_convert_row_columnar_dataset(
        update_data, ComponentAttributeFilterOptions.relevant, DatasetType.update
    )
    with pytest.raises(PowerGridError, match="The id cannot be found:"):
        model.update(update_data=update_data_col)


def test_copy_model(model: PowerGridModel, sym_output):
    model_2 = copy(model)
    result = model_2.calculate_power_flow()
    compare_result(result, sym_output, rtol=0.0, atol=1e-8)


def test_deepcopy_model(model: PowerGridModel, empty_model: PowerGridModel, sym_output, update_batch, sym_output_batch):
    # list containing different models twice
    model_list = [model, empty_model, model, empty_model]

    new_model_list = deepcopy(model_list)

    # check if identities are as expected
    assert id(new_model_list[0]) != id(model_list[0])
    assert id(new_model_list[1]) != id(model_list[1])
    assert id(new_model_list[0]) != id(new_model_list[1])
    assert id(new_model_list[0]) == id(new_model_list[2])
    assert id(new_model_list[1]) == id(new_model_list[3])

    # check if the deepcopied objects are really independent from the original ones
    # by modifying the copies and seeing if the original one is impacted by this change
    new_model_list[0].update(update_data=get_dataset_scenario(update_batch, 0))

    new_expected_result = get_dataset_scenario(sym_output_batch, 0)
    new_result_0 = new_model_list[0].calculate_power_flow()
    compare_result(new_result_0, new_expected_result, rtol=0.0, atol=1e-8)
    # at index 0 and 2 should be the same objects, check if changing the object at index 0
    # and obtaining a power flow result is ident to the result at index 2
    new_result_2 = new_model_list[2].calculate_power_flow()
    compare_result(new_result_2, new_expected_result, rtol=0.0, atol=1e-8)

    result = model.calculate_power_flow()
    compare_result(result, sym_output, rtol=0.0, atol=1e-8)


def test_repr_and_str(model: PowerGridModel, empty_model: PowerGridModel):
    repr_empty_model_expected = "PowerGridModel (0 components)\n"
    assert repr_empty_model_expected == repr(empty_model)
    assert repr_empty_model_expected == str(empty_model)

    repr_model_expected = "PowerGridModel (3 components)\n  - node: 1\n  - source: 1\n  - sym_load: 1\n"
    assert repr_model_expected == repr(model)
    assert repr_model_expected == str(model)


def test_get_indexer(model: PowerGridModel):
    ids = np.array([2, 2])
    expected_indexer = np.array([0, 0])
    indexer = model.get_indexer(ComponentType.sym_load, ids)
    np.testing.assert_allclose(expected_indexer, indexer)


def test_batch_power_flow(model: PowerGridModel, update_batch: BatchDataset, sym_output_batch):
    result = model.calculate_power_flow(update_data=update_batch)
    compare_result(result, sym_output_batch, rtol=0.0, atol=1e-8)


def test_construction_error(input):
    input[ComponentType.sym_load]["id"][0] = 0
    with pytest.raises(PowerGridError, match="Conflicting id detected:"):
        PowerGridModel(input)


def test_single_calculation_error(model: PowerGridModel):
    with pytest.raises(IterationDiverge, match="Iteration failed to converge after"):
        model.calculate_power_flow(max_iterations=1, error_tolerance=1e-100)
    with pytest.raises(InvalidCalculationMethod, match="The calculation method is invalid for this calculation!"):
        model.calculate_state_estimation(calculation_method="iterative_current")

    for calculation_method in ("linear", "newton_raphson", "iterative_current", "linear_current", "iterative_linear"):
        with pytest.raises(InvalidCalculationMethod):
            model.calculate_short_circuit(calculation_method=calculation_method)


def test_batch_calculation_error(model: PowerGridModel, update_batch):
    # wrong id
    update_batch[ComponentType.sym_load]["data"]["id"][1] = 5
    # with error
    with pytest.raises(PowerGridBatchError) as e:
        model.calculate_power_flow(update_data=update_batch)
    error = e.value
    np.testing.assert_allclose(error.failed_scenarios, [1])
    np.testing.assert_allclose(error.succeeded_scenarios, [0])
    assert "The id cannot be found:" in error.error_messages[0]


def test_batch_calculation_error_continue(model: PowerGridModel, update_batch, sym_output_batch):
    # wrong id
    update_batch[ComponentType.sym_load]["data"]["id"][1] = 5
    result = model.calculate_power_flow(update_data=update_batch, continue_on_batch_error=True)
    # assert error
    error = model.batch_error
    assert error is not None
    np.testing.assert_allclose(error.failed_scenarios, [1])
    np.testing.assert_allclose(error.succeeded_scenarios, [0])
    assert "The id cannot be found:" in error.error_messages[0]
    # assert value result for scenario 0
    result = {ComponentType.node: result[ComponentType.node][error.succeeded_scenarios, :]}
    expected_result = {ComponentType.node: sym_output_batch[ComponentType.node][error.succeeded_scenarios, :]}
    compare_result(result, expected_result, rtol=0.0, atol=1e-8)
    # general error before the batch
    with pytest.raises(PowerGridError, match="The calculation method is invalid for this calculation!"):
        model.calculate_state_estimation(
            calculation_method="iterative_current",
            update_data={
                ComponentType.source: initialize_array(DatasetType.update, ComponentType.source, shape=(5, 0))
            },
            continue_on_batch_error=True,
        )


def test_empty_input():
    node = initialize_array(DatasetType.input, ComponentType.node, 0)
    line = initialize_array(DatasetType.input, ComponentType.line, 0)
    sym_load = initialize_array(DatasetType.input, ComponentType.sym_load, 0)
    source = initialize_array(DatasetType.input, ComponentType.source, 0)

    input_data = {
        ComponentType.node: node,
        ComponentType.line: line,
        ComponentType.sym_load: sym_load,
        ComponentType.source: source,
    }

    assert_valid_input_data(input_data)
    model = PowerGridModel(input_data, system_frequency=50.0)

    result = model.calculate_power_flow()

    assert result == {}


@pytest.fixture
def input_sym_load_col(input_row):
    return compatibility_convert_row_columnar_dataset(
        input_row,
        {
            ComponentType.node: None,
            ComponentType.source: None,
            ComponentType.sym_load: ComponentAttributeFilterOptions.relevant,
        },
        DatasetType.input,
    )


@pytest.fixture(params=[pytest.param("input_row", id="input_row"), pytest.param("input_sym_load_col", id="input_col")])
def minimal_input(request):
    return request.getfixturevalue(request.param)


def update_sym_load_row():
    sym_load = initialize_array(DatasetType.update, ComponentType.sym_load, (2, 1))
    sym_load["id"] = [[2], [2]]
    sym_load["q_specified"] = [[100.0], [300.0]]
    return {ComponentType.sym_load: sym_load}


def update_sym_load_row_optional_id():
    sym_load = initialize_array(DatasetType.update, ComponentType.sym_load, (2, 1))
    sym_load["q_specified"] = [[100.0], [300.0]]
    return {ComponentType.sym_load: sym_load}


def update_sym_load_row_invalid_id():
    sym_load = initialize_array(DatasetType.update, ComponentType.sym_load, (2, 1))
    sym_load["id"] = [[2], [5]]
    sym_load["q_specified"] = [[100.0], [300.0]]
    return {ComponentType.sym_load: sym_load}


def update_sym_load_col(update_sym_load_row):
    return compatibility_convert_row_columnar_dataset(
        update_sym_load_row, ComponentAttributeFilterOptions.relevant, DatasetType.update
    )


def update_sym_load_sparse(update_data):
    return {
        ComponentType.sym_load: {
            "data": update_data[ComponentType.sym_load].reshape(-1),
            "indptr": np.array([0, 1, 2]),
        },
    }


@pytest.mark.parametrize(
    "minimal_update",
    [
        pytest.param(update_sym_load_row(), id="update_dense_row"),
        pytest.param(update_sym_load_col(update_sym_load_row()), id="update_dense_col"),
        pytest.param(update_sym_load_sparse(update_sym_load_row()), id="update_sparse_row"),
        pytest.param(update_sym_load_col(update_sym_load_sparse(update_sym_load_row())), id="update_sparse_col"),
    ],
)
def test_update_ids_batch(minimal_update, minimal_input):
    output_data = PowerGridModel(minimal_input).calculate_power_flow(update_data=minimal_update)
    np.testing.assert_almost_equal(output_data[ComponentType.node]["u"], np.array([[90.0], [70.0]]))


@pytest.mark.parametrize(
    "minimal_update",
    [
        pytest.param(update_sym_load_row_optional_id(), id="update_dense_row"),
        pytest.param(update_sym_load_col(update_sym_load_row_optional_id()), id="update_dense_col"),
        pytest.param(update_sym_load_sparse(update_sym_load_row_optional_id()), id="update_sparse_row"),
        pytest.param(
            update_sym_load_col(update_sym_load_sparse(update_sym_load_row_optional_id())), id="update_sparse_col"
        ),
    ],
)
def test_update_id_optional(minimal_update, minimal_input):
    output_data = PowerGridModel(minimal_input).calculate_power_flow(update_data=minimal_update)
    np.testing.assert_almost_equal(output_data[ComponentType.node]["u"], np.array([[90.0], [70.0]]))


def test_update_id_mixed(minimal_input):
    update_sym_load_no_id = initialize_array(DatasetType.update, ComponentType.sym_load, (3, 1))
    update_sym_load_no_id["p_specified"] = [[30e6], [15e5], [0]]

    update_source_indptr = np.array([0, 1, 1, 2])
    update_source = initialize_array(DatasetType.update, ComponentType.source, 2)
    update_source["id"] = 1
    update_source["status"] = 0

    update_batch = {
        ComponentType.sym_load: update_sym_load_no_id,
        ComponentType.source: {"indptr": update_source_indptr, "data": update_source},
    }

    _output_data = PowerGridModel(minimal_input).calculate_power_flow(update_data=update_batch)


@pytest.mark.parametrize(
    "minimal_update",
    [
        update_sym_load_row_invalid_id(),
        update_sym_load_col(update_sym_load_row_invalid_id()),
        update_sym_load_sparse(update_sym_load_row_invalid_id()),
        update_sym_load_col(update_sym_load_sparse(update_sym_load_row_invalid_id())),
    ],
)
def test_update_id_error(minimal_update, minimal_input):
    with pytest.raises(PowerGridBatchError) as e:
        PowerGridModel(minimal_input).calculate_power_flow(update_data=minimal_update)
    assert e.value.failed_scenarios == [1]
    assert "The id cannot be found: 5" in e.value.error_messages[0]



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

# define dataclass for meta data

from dataclasses import dataclass

from dataclasses_json import DataClassJsonMixin


@dataclass
class Attribute(DataClassJsonMixin):
    data_type: str
    names: str | list[str]
    description: str
    nan_value: str | None = None


@dataclass
class AttributeClass(DataClassJsonMixin):
    name: str
    attributes: list[Attribute]
    full_attributes: list[Attribute] | None = None
    base: str | None = None
    is_template: bool = False
    full_name: str | None = None
    specification_names: list[str] | None = None
    base_attributes: dict[str, list[Attribute]] | None = None


@dataclass
class DatasetMetaData(DataClassJsonMixin):
    name: str
    include_guard: str
    classes: list[AttributeClass]


@dataclass
class ObjectMapData(DataClassJsonMixin):
    names: list[str]
    class_name: str


@dataclass
class DatasetMapData(DataClassJsonMixin):
    name: str
    is_template: bool
    components: list[ObjectMapData]


@dataclass
class AllDatasetMapData(DataClassJsonMixin):
    all_datasets: list[DatasetMapData]



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""
Power grid model buffer handler tests
"""

import numpy as np
import pytest

from power_grid_model._core.buffer_handling import (
    _get_raw_attribute_data_view,
    get_buffer_properties,
    get_buffer_view,
)
from power_grid_model._core.dataset_definitions import ComponentType, DatasetType
from power_grid_model._core.power_grid_meta import initialize_array, power_grid_meta_data

SINGLE_DATASET_NDIM = 1
BATCH_DATASET_NDIM = 2
SCENARIO_TOTAL_ELEMENTS = 4
BATCH_TOTAL_ELEMENTS = 8


def load_data(component_type, is_batch, is_sparse, is_columnar):
    """Creates load data of different formats for testing"""
    shape = (BATCH_DATASET_NDIM, SCENARIO_TOTAL_ELEMENTS) if is_batch else (SCENARIO_TOTAL_ELEMENTS,)
    load = initialize_array(DatasetType.update, component_type, shape)
    columnar_names = ["p_specified", "q_specified"]

    if is_columnar:
        if is_sparse:
            return {
                "indptr": np.array([0, 5, BATCH_TOTAL_ELEMENTS]),
                "data": {k: load.reshape(-1)[k] for k in columnar_names},
            }
        return {k: load[k] for k in columnar_names}
    if is_sparse:
        return {"indptr": np.array([0, 5, BATCH_TOTAL_ELEMENTS]), "data": load.reshape(-1)}
    return load


@pytest.mark.parametrize(
    ("component_type", "is_batch", "is_columnar"),
    [
        pytest.param(ComponentType.sym_load, True, True, id="sym_load-batch-columnar"),
        pytest.param(ComponentType.sym_load, True, False, id="sym_load-batch-row_based"),
        pytest.param(ComponentType.sym_load, False, True, id="sym_load-single-columnar"),
        pytest.param(ComponentType.sym_load, False, False, id="sym_load-single-row_based"),
        pytest.param(ComponentType.asym_load, True, True, id="asym_load-batch-columnar"),
        pytest.param(ComponentType.asym_load, True, False, id="asym_load-batch-row_based"),
        pytest.param(ComponentType.asym_load, False, True, id="asym_load-single-columnar"),
        pytest.param(ComponentType.asym_load, False, False, id="asym_load-single-row_based"),
    ],
)
def test__get_dense_buffer_properties(component_type, is_batch, is_columnar):
    data = load_data(component_type, is_batch=is_batch, is_columnar=is_columnar, is_sparse=False)
    schema = power_grid_meta_data[DatasetType.update][component_type]
    batch_size = BATCH_DATASET_NDIM if is_batch else None
    properties = get_buffer_properties(data, schema=schema, is_batch=is_batch, batch_size=batch_size)

    assert not properties.is_sparse
    assert properties.is_batch == is_batch
    assert properties.batch_size == (BATCH_DATASET_NDIM if is_batch else SINGLE_DATASET_NDIM)
    assert properties.n_elements_per_scenario == SCENARIO_TOTAL_ELEMENTS
    assert properties.n_total_elements == BATCH_TOTAL_ELEMENTS if is_batch else 4
    if is_columnar:
        assert properties.columns == list(data.keys())
    else:
        assert properties.columns is None


@pytest.mark.parametrize(
    ("component_type", "is_columnar"),
    [
        pytest.param(ComponentType.sym_load, True, id="sym_load-columnar"),
        pytest.param(ComponentType.sym_load, False, id="sym_load-row_based"),
        pytest.param(ComponentType.asym_load, True, id="asym_load-columnar"),
        pytest.param(ComponentType.asym_load, False, id="asym_load-row_based"),
    ],
)
def test__get_sparse_buffer_properties(component_type, is_columnar):
    data = load_data(component_type, is_batch=True, is_columnar=is_columnar, is_sparse=True)

    schema = power_grid_meta_data[DatasetType.update][component_type]
    properties = get_buffer_properties(data, schema=schema, batch_size=2)

    assert properties.is_sparse
    assert properties.is_batch
    assert properties.batch_size == BATCH_DATASET_NDIM
    assert properties.n_elements_per_scenario == -1
    assert properties.n_total_elements == BATCH_TOTAL_ELEMENTS
    if is_columnar:
        assert properties.columns == list(data["data"].keys())
    else:
        assert properties.columns is None


@pytest.mark.parametrize(
    ("component_type", "is_columnar"),
    [
        pytest.param(ComponentType.sym_load, True, id="sym_load-single-columnar"),
        pytest.param(ComponentType.sym_load, False, id="sym_load-single-row_based"),
        pytest.param(ComponentType.asym_load, True, id="asym_load-single-columnar"),
        pytest.param(ComponentType.asym_load, False, id="asym_load-single-row_based"),
    ],
)
def test__get_buffer_properties__batch_requested_for_single_data(component_type, is_columnar):
    data = load_data(component_type, is_batch=False, is_columnar=is_columnar, is_sparse=False)
    schema = power_grid_meta_data[DatasetType.update][component_type]

    with pytest.raises(
        ValueError, match=r"Incorrect\/inconsistent data provided: single data provided but batch data expected"
    ):
        get_buffer_properties(data, schema=schema, is_batch=True, batch_size=BATCH_DATASET_NDIM)


@pytest.mark.parametrize(
    ("component_type", "is_sparse", "is_columnar"),
    [
        pytest.param(ComponentType.sym_load, False, True, id="sym_load-batch-columnar"),
        pytest.param(ComponentType.sym_load, False, False, id="sym_load-batch-row_based"),
        pytest.param(ComponentType.asym_load, False, True, id="asym_load-batch-columnar"),
        pytest.param(ComponentType.asym_load, False, False, id="asym_load-batch-row_based"),
        pytest.param(ComponentType.sym_load, True, True, id="sym_load-columnar"),
        pytest.param(ComponentType.sym_load, True, False, id="sym_load-row_based"),
        pytest.param(ComponentType.asym_load, True, True, id="asym_load-columnar"),
        pytest.param(ComponentType.asym_load, True, False, id="asym_load-row_based"),
    ],
)
def test__get_buffer_properties__single_requested_for_batch(component_type, is_sparse, is_columnar):
    data = load_data(component_type, is_batch=True, is_columnar=is_columnar, is_sparse=is_sparse)
    schema = power_grid_meta_data[DatasetType.update][component_type]

    if is_sparse:
        with pytest.raises(ValueError, match="Sparse data must be batch data"):
            get_buffer_properties(data, schema=schema, is_batch=False, batch_size=None)
    else:
        with pytest.raises(
            ValueError, match=r"Incorrect\/inconsistent data provided: batch data provided but single data expected"
        ):
            get_buffer_properties(data, schema=schema, is_batch=False, batch_size=None)


@pytest.mark.parametrize(
    ("component_type", "is_sparse", "is_columnar"),
    [
        pytest.param(ComponentType.sym_load, False, True, id="sym_load-batch-columnar"),
        pytest.param(ComponentType.sym_load, False, False, id="sym_load-batch-row_based"),
        pytest.param(ComponentType.asym_load, False, True, id="asym_load-batch-columnar"),
        pytest.param(ComponentType.asym_load, False, False, id="asym_load-batch-row_based"),
        pytest.param(ComponentType.sym_load, True, True, id="sym_load-columnar"),
        pytest.param(ComponentType.sym_load, True, False, id="sym_load-row_based"),
        pytest.param(ComponentType.asym_load, True, True, id="asym_load-columnar"),
        pytest.param(ComponentType.asym_load, True, False, id="asym_load-row_based"),
    ],
)
def test__get_buffer_properties__wrong_batch_size(component_type, is_sparse, is_columnar):
    data = load_data(component_type, is_batch=True, is_columnar=is_columnar, is_sparse=is_sparse)
    schema = power_grid_meta_data[DatasetType.update][component_type]

    actual_batch_size = BATCH_DATASET_NDIM
    wrong_batch_size = actual_batch_size + 1

    with pytest.raises(
        ValueError,
        match=(
            f"Incorrect/inconsistent batch size provided: {actual_batch_size} scenarios provided but "
            f"{wrong_batch_size} scenarios expected."
        ),
    ):
        get_buffer_properties(data, schema=schema, is_batch=True, batch_size=wrong_batch_size)


@pytest.mark.parametrize(
    ("component", "is_batch", "is_columnar", "is_sparse"),
    [
        pytest.param(ComponentType.sym_load, True, True, False, id="sym_load-columnar"),
        pytest.param(ComponentType.asym_load, True, True, False, id="asym_load-columnar"),
    ],
)
def test__get_raw_attribute_data_view(component, is_batch, is_columnar, is_sparse):
    schema = power_grid_meta_data[DatasetType.update][component]

    data = load_data(
        component_type=component,
        is_batch=is_batch,
        is_columnar=is_columnar,
        is_sparse=is_sparse,
    )
    asym_dense_batch_last_dim = 3
    if component == ComponentType.asym_load:
        assert data["p_specified"].shape[-1] == asym_dense_batch_last_dim

    buffer_view = get_buffer_view(
        data,
        schema=schema,
        is_batch=is_batch,
    )
    assert buffer_view is not None
    assert buffer_view.batch_size == BATCH_DATASET_NDIM
    assert buffer_view.n_elements_per_scenario == SCENARIO_TOTAL_ELEMENTS
    assert buffer_view.total_elements == BATCH_TOTAL_ELEMENTS


@pytest.mark.parametrize(
    ("component", "attribute"),
    [
        pytest.param(ComponentType.asym_load, "p_specified", id="asym_load-shape_missmatch"),
    ],
)
def test__get_raw_attribute_data_view_fail(component, attribute):
    schema = power_grid_meta_data[DatasetType.update][component]

    data = load_data(
        component_type=component,
        is_batch=True,
        is_columnar=True,
        is_sparse=False,
    )

    old_shape = data[attribute].shape
    new_shape = list(old_shape)
    # because in _get_raw_attribute_data_view we check if the dimension of the last element is not 3 to raise an error
    new_shape[-1] = 2
    data[attribute] = np.zeros(new_shape, dtype=data[attribute].dtype)

    asym_dense_batch_last_dim = 3
    unsupported_asym_dense_batch_last_dim = 2
    updated_shape = data[attribute].shape
    assert old_shape != updated_shape
    assert old_shape[-1] == asym_dense_batch_last_dim
    assert updated_shape[-1] == unsupported_asym_dense_batch_last_dim

    with pytest.raises(ValueError, match="Given data has a different schema than supported"):
        get_buffer_view(data, schema=schema, is_batch=True)


@pytest.mark.parametrize(
    ("component", "is_batch", "is_columnar", "is_sparse"),
    [
        pytest.param(ComponentType.sym_load, True, True, False, id="sym_load"),
        pytest.param(ComponentType.asym_load, True, True, False, id="asym_load-columnar"),
    ],
)
def test__get_raw_attribute_data_view_direct(component, is_batch, is_columnar, is_sparse):
    attribute = "p_specified"
    schema = power_grid_meta_data[DatasetType.update][component]

    data = load_data(
        component_type=component,
        is_batch=is_batch,
        is_columnar=is_columnar,
        is_sparse=is_sparse,
    )

    attribute_data = data[attribute]

    _get_raw_attribute_data_view(attribute_data, schema, "p_specified")


@pytest.mark.parametrize(
    ("component", "attr_data_shape", "attribute"),
    [
        pytest.param(ComponentType.asym_load, (2, 4, 3), "p_specified", id="asym_load-columnar"),
        pytest.param(ComponentType.asym_load, (1, 4, 3), "p_specified", id="asym_load-columnar"),
        pytest.param(ComponentType.asym_load, (2, 5, 3), "p_specified", id="asym_load-columnar"),
        pytest.param(ComponentType.asym_load, (2, 6, 3), "p_specified", id="asym_load-columnar"),
        pytest.param(ComponentType.asym_load, 3, "p_specified", id="asym_load-columnar"),
        pytest.param(ComponentType.asym_load, (3, 3), "p_specified", id="asym_load-columnar"),
        pytest.param(ComponentType.asym_load, (2, 4, 3), "q_specified", id="asym_load-columnar"),
        pytest.param(ComponentType.asym_load, (1, 4, 3), "q_specified", id="asym_load-columnar"),
        pytest.param(ComponentType.asym_load, (2, 5, 3), "q_specified", id="asym_load-columnar"),
        pytest.param(ComponentType.asym_load, (2, 6, 3), "q_specified", id="asym_load-columnar"),
        pytest.param(ComponentType.asym_load, 3, "q_specified", id="asym_load-columnar"),
        pytest.param(ComponentType.asym_load, (3, 3), "q_specified", id="asym_load-columnar"),
        # sym component
        pytest.param(ComponentType.sym_load, 3, "p_specified", id="sym_load-columnar"),
        pytest.param(ComponentType.sym_load, (3, 3), "p_specified", id="sym_load-columnar"),
        pytest.param(ComponentType.sym_load, (4, 3), "p_specified", id="sym_load-columnar"),
        pytest.param(ComponentType.sym_load, (4, 0), "p_specified", id="sym_load-columnar"),
        pytest.param(ComponentType.sym_load, 0, "p_specified", id="sym_load-columnar"),
        pytest.param(ComponentType.sym_load, 3, "q_specified", id="sym_load-columnar"),
        pytest.param(ComponentType.sym_load, (3, 3), "q_specified", id="sym_load-columnar"),
        pytest.param(ComponentType.sym_load, (4, 3), "q_specified", id="sym_load-columnar"),
        pytest.param(ComponentType.sym_load, (4, 0), "q_specified", id="sym_load-columnar"),
        pytest.param(ComponentType.sym_load, (0, 4), "q_specified", id="sym_load-columnar"),
        pytest.param(ComponentType.sym_load, 0, "q_specified", id="sym_load-columnar"),
    ],
)
def test__get_raw_attribute_data_view_directly(component, attr_data_shape, attribute):
    arr = np.zeros(attr_data_shape)
    schema = power_grid_meta_data[DatasetType.update][component]

    _get_raw_attribute_data_view(arr, schema, attribute)


@pytest.mark.parametrize(
    ("component", "attr_data_shape", "attribute"),
    [
        pytest.param(ComponentType.asym_load, (2, 4, 4), "p_specified", id="asym_load-columnar"),
        pytest.param(ComponentType.asym_load, (2, 6, 4), "p_specified", id="asym_load-columnar"),
        pytest.param(ComponentType.asym_load, 0, "p_specified", id="asym_load-columnar"),
        pytest.param(ComponentType.asym_load, (2, 0, 0), "p_specified", id="asym_load-columnar"),
        pytest.param(ComponentType.asym_load, (2, 4, 0), "p_specified", id="asym_load-columnar"),
        pytest.param(ComponentType.asym_load, (2, 4, 4), "q_specified", id="asym_load-columnar"),
        pytest.param(ComponentType.asym_load, (2, 6, 4), "q_specified", id="asym_load-columnar"),
        pytest.param(ComponentType.asym_load, 0, "q_specified", id="asym_load-columnar"),
        pytest.param(ComponentType.asym_load, (2, 0, 0), "q_specified", id="asym_load-columnar"),
        pytest.param(ComponentType.asym_load, (2, 4, 0), "q_specified", id="asym_load-columnar"),
        # sym component
        pytest.param(ComponentType.sym_load, (2, 4, 4), "p_specified", id="sym_load-columnar"),
        pytest.param(ComponentType.sym_load, (2, 4, 3), "p_specified", id="sym_load-columnar"),
        pytest.param(ComponentType.sym_load, (0, 0, 1), "p_specified", id="sym_load-columnar"),
        pytest.param(ComponentType.sym_load, (2, 4, 4), "q_specified", id="sym_load-columnar"),
        pytest.param(ComponentType.sym_load, (2, 4, 3), "q_specified", id="sym_load-columnar"),
        pytest.param(ComponentType.sym_load, (0, 0, 1), "q_specified", id="sym_load-columnar"),
    ],
)
def test__get_raw_attribute_data_view_directly_fail(component, attr_data_shape, attribute):
    arr = np.zeros(attr_data_shape)
    schema = power_grid_meta_data[DatasetType.update][component]

    with pytest.raises(ValueError, match="Given data has a different schema than supported"):
        _get_raw_attribute_data_view(arr, schema, attribute)
